from mock import call, patch
import os
from pkg_resources import resource_filename
import requests
import shutil
import subprocess
import tempfile
from unittest import TestCase
from six.moves.urllib.parse import urlparse

from codetree.handlers import charm
from codetree.handlers import CharmStoreHandler
from codetree.handlers.utils import handler_types
from codetree.handlers.exceptions import InvalidOption


class MockedReq(object):
    """ A class to mock out a request object without actually hitting the internet.
    """

    def __init__(self, file_data=None, sha=None):

        self.test_file = resource_filename("tests", "/test_data/charm.zip")
        if file_data is None:
            with open(self.test_file, 'rb') as test_zip:
                self.file_data = test_zip.read()
        else:
            self.file_data = file_data
        if sha is None:
            self._headers = {'Content-Sha384':
                             '5a1481850e29647059f21f8bf23f082649d3f5a9808481fe122db41f7a827765025da5e17da9e9ae522140113b2f1804'}
        else:
            self._headers = {'Content-Sha384': sha}

    @property
    def headers(self):
        return self._headers

    def iter_content(self, size=None):
        yield self.file_data
        raise StopIteration

    def json(self):
        return {'Id': 'cs:trusty/charm'}


class TestCharmStoreHandler(TestCase):
    def setUp(self):
        charm._have_charm_tools = lambda: False
        self.cs = CharmStoreHandler('cs:trusty/charm')
        self.get_url = "{}/{}/archive".format(self.cs.base_url, 'trusty/charm')
        self.tdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tdir)

    def test_url_handling(self):
        urls = (
            'cs:trusty/auditd',
            'cs:trusty/auditd-0'
        )
        for url in urls:
            self.assertIn(urlparse(url).scheme, CharmStoreHandler.schemes)

        assert 'cs' in handler_types().keys()

    def test_metadata(self):
        dest = os.path.join(self.tdir, 'test_download')
        os.mkdir(dest, 0o0755)
        self.cs.charmstore_url = 'cs:trusty/test-auditd-1'

        self.assertEqual(None, self.cs._get_url(dest))

        self.cs._set_url(dest)
        self.assertEqual('cs:trusty/test-auditd-1;channel=stable', self.cs._get_url(dest))

    @patch("codetree.handlers.charm.logging.error")
    @patch("codetree.handlers.charm.logging.info")
    @patch("codetree.handlers.charm.requests.get")
    def test_download(self, _get, _logging_info, _logging_error):
        dest = os.path.join(self.tdir, 'test_download')
        _get.return_value = MockedReq()

        assert self.cs.get(dest, {})
        assert _get.call_count == 2
        expected_request_args = [call('{}/trusty/charm/meta/id?channel=stable'.format(self.cs.base_url)),
                                 call('{}/trusty/charm/archive'.format(self.cs.base_url))]
        assert _get.call_args_list == expected_request_args
        expected_logging = [call("Downloading {} from charm store to {}".format(self.cs.source, dest)),
                            call("{} charm retrieved from {}".format(dest, self.cs.charmstore_url))]
        self.assertEqual(_logging_info.call_args_list, expected_logging)
        _get.reset_mock()
        _logging_info.reset_mock()

        # check only, no update needed
        assert self.cs.get(dest, {})
        assert _get.call_count == 1
        expected_request_args = [call('{}/trusty/charm/meta/id?channel=stable'.format(self.cs.base_url))]
        assert _get.call_args_list == expected_request_args
        expected_logging = [call("{} is up to date from url {}, skipping.".format(dest, self.cs.charmstore_url))]
        assert _logging_info.call_args_list == expected_logging
        _get.reset_mock()
        _logging_info.reset_mock()

        # Edit the metadata forcing a re-download
        self.cs.charmstore_url += 'test'
        self.cs._set_url(dest)
        assert self.cs.get(dest, {})
        assert _get.call_count == 2
        expected_request_args = [call('{}/trusty/charm/meta/id?channel=stable'.format(self.cs.base_url)),
                                 call('{}/trusty/charm/archive'.format(self.cs.base_url))]
        assert _get.call_args_list == expected_request_args
        expected_logging = [call("Downloading {} from charm store to {}".format(self.cs.source, dest)),
                            call("{} charm retrieved from {}".format(dest, self.cs.charmstore_url))]
        assert _logging_info.call_args_list == expected_logging
        _get.reset_mock()
        _logging_info.reset_mock()

        # Trick it into thinking it's a git repo, `get()` should refuse to
        # update and log an error.
        os.remove(os.path.join(dest, self.cs.metadata_filename))
        with open(os.path.join(dest, '.git'), 'w') as git:
            git.write('This is a git repo')
        # Confirm we return False which will be considered an error.
        self.assertFalse(self.cs.get(dest, {}))
        self.assertEqual(_get.call_count, 1)
        expected_request_args = [call('{}/trusty/charm/meta/id?channel=stable'.format(self.cs.base_url))]
        assert _get.call_args_list == expected_request_args
        expected_logging = [call("Failed to update {} to {} because it looks like a bzr or git branch, and the "
                                 "'overwrite' option wasn't set".format(dest, self.cs.charmstore_url))]
        assert _logging_error.call_args_list == expected_logging

    @patch("codetree.handlers.charm.logging.info")
    @patch("codetree.handlers.charm.requests.get")
    def test_download_with_overwrite(self, _get, _logging):
        dest = os.path.join(self.tdir, 'test_overwrite')
        _get.return_value = MockedReq()

        assert self.cs.get(dest, {'overwrite': True})
        _get.assert_called_with(self.get_url)
        expected_logging = [call("Downloading {} from charm store to {}".format(self.cs.source, dest)),
                            call("{} charm retrieved from trusty/charm".format(dest))]
        self.assertEqual(_logging.call_args_list, expected_logging)
        _logging.reset_mock()

        # Now let's trick it into thinking it's a git repo
        with open(os.path.join(dest, '.git'), 'w') as git:
            git.write('This is a git repo')

        # Second run should overwrite (redownload) even though it's a git branch
        assert self.cs.get(dest, {'overwrite': True})
        _get.assert_called_with(self.get_url)

        expected_logging = [call("Downloading {} from charm store to {}".format(self.cs.source, dest)),
                            call("{} charm retrieved from {}".format(dest, self.cs.charmstore_url))]
        assert _logging.call_args_list == expected_logging

    @patch("codetree.handlers.charm.logging.info")
    @patch("codetree.handlers.charm.requests.get")
    def test_download_from_channel(self, _get, _logging):
        dest = os.path.join(self.tdir, 'test_download')
        _get.return_value = MockedReq()

        assert self.cs.get(dest, {'channel': 'edge'})
        assert _get.call_count == 2
        expected_request_args = [call('{}/trusty/charm/meta/id?channel=edge'.format(self.cs.base_url)),
                                 call('{}/trusty/charm/archive'.format(self.cs.base_url))]
        assert _get.call_args_list == expected_request_args
        expected_logging = [call("Downloading {} from charm store to {}".format(self.cs.source, dest)),
                            call("{} charm retrieved from {}".format(dest, self.cs.charmstore_url))]
        assert _logging.call_args_list == expected_logging
        _get.reset_mock()
        _logging.reset_mock()

        # Second run should re-download as we have no way of knowing if the upstream is the same as on disk
        self.assertRaises(InvalidOption, self.cs.get, dest, {'channel': 'fakechannel'})
        _get.reset_mock()
        _logging.reset_mock()

    @patch("codetree.handlers.charm.requests.get")
    def test_failed_sha(self, _get):
        dest = os.path.join(self.tdir, 'test_failed_sha')
        _get.return_value = MockedReq(sha='toosmallforasha')

        with self.assertRaises(requests.HTTPError):
            self.cs.get(dest, {})

    @patch("codetree.handlers.charm.requests.get")
    def test_download_invalid_zip(self, _get):
        dest = os.path.join(self.tdir, 'test_invalid_zip')

        test_sha = '0291dbcaaa51039c59ae631f106d95791fb8c74815fee44f3f31fea7f9f30d001e24d038e90ed632d1d71a6f4082ad5c'
        _get.return_value = MockedReq(file_data=b'not a valid zip', sha=test_sha)

        with self.assertRaises(subprocess.CalledProcessError):
            self.cs.get(dest, {})


class TestCharmStoreHandlerUsingCharm(TestCase):
    def setUp(self):
        charm._have_charm_tools = lambda: True
        self.cs = CharmStoreHandler('cs:trusty/charm')
        self.tdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tdir)

    @patch("codetree.handlers.charm.logging.info")
    @patch("codetree.handlers.charm.subprocess.check_output")
    def test_download(self, _check_output, _logging):
        dest = os.path.join(self.tdir, 'test_download')

        def simulate_charm(cmd, *args, **kwargs):
            if cmd[1] == 'show':
                return '{"id": {"Id": "cs:trusty/charm-10"}}'
            elif cmd[1] == 'pull':
                os.makedirs(cmd[5])

        _check_output.side_effect = simulate_charm

        assert self.cs.get(dest, {})
        assert _check_output.call_count == 2
        expected_request_args = [
            call(('charm', 'show', '--format=json', '--channel', 'stable', 'cs:trusty/charm', 'id'), stderr=subprocess.STDOUT, universal_newlines=True),
            call(('charm', 'pull', '--channel', 'stable', 'cs:trusty/charm-10', dest), stderr=subprocess.STDOUT, universal_newlines=True)]
        assert _check_output.call_args_list == expected_request_args
        expected_logging = [call("Downloading cs:trusty/charm-10 from charm store to {}".format(dest)),
                            call("{} charm retrieved from {}".format(dest, self.cs.charmstore_url))]
        self.assertEqual(_logging.call_args_list, expected_logging)
